#!/bin/sh
##########################################################################
#                                                                        #
#                         generate_dependencies.sh                       #
#                                                                        #
##########################################################################
#                                                                        #
# A script for generating the dependencies of .as files                  #
#                                                                        #
#                                                                        #
##########################################################################
#                                                                        #
# Copyright (C) 2005  Research Institute for Symbolic Computation, J.    #
# Kepler University, Linz, Austria                                       #
#                                                                        #
# Written by Christian Aistleitner                                       #
#                                                                        #
# This program is free software; you can redistribute it and/or          #
# modify it under the terms of the GNU General Public License version 2, #
# as published by the Free Software Foundation.                          #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program; if not, write to the Free Software            #
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,             #
# MA  02110-1301, USA.                                                   #
#                                                                        #
##########################################################################

ERROROK=0
ERRORUNDEFINEDERROR=100
ERRORARGUMENTS=$(($ERRORUNDEFINEDERROR+1))

TRUE=0
FALSE=1

##########################################################################
#                                                                        #
# function: generate_file_variables                                      #
#                                                                        #
##########################################################################
#                                                                        #
# generates the OFILE, AOFILE variables for a file.                      #
#                                                                        #
##########################################################################
#                                                                        #
# input:                                                                 #
#                                                                        #
#    THISIDX             - the index of the file in the arrays           #
#    ASFILE[$THISIDX]    - the .as file                                  #
#    LIBNAME[$THISIDX]   - the relative library name without .a or .al   #
#                                                                        #
##########################################################################
#                                                                        #
# output:                                                                #
#                                                                        #
#    AOFILE[$THISIDX]    - the .ao file                                  #
#    OFILE[$THISIDX]     - the .o  file                                  #
#    LIBAOFILE[$THISIDX] - the .ao file in the library                   #
#    LIBOFILE[$THISIDX]  - the .o  file in the library                   #
#    INDEPENDENTFILE[$THISIDX]  - this file does not depend on other     #
#                          files, and might start a library              #
#                                                                        #
##########################################################################

function generate_file_variables ()
{
  AOFILE[$THISIDX]=`echo ${ASFILE[$THISIDX]} | sed -e 's/\(.*\/\)\([^\/]*.\).as$/\2\.ao/'`
  OFILE[$THISIDX]=`echo ${AOFILE[$THISIDX]} | sed -e 's/ao$/o/'`
  LIBAOFILE[$THISIDX]="${LIBNAME[$THISIDX]}\$(LIBRARYVARIANTSUFFIX).al(${AOFILE[$THISIDX]})"
  LIBOFILE[$THISIDX]="${LIBNAME[$THISIDX]}\$(LIBRARYVARIANTSUFFIX).a(${OFILE[$THISIDX]})"
  INDEPENDENTFILE[$THISIDX]=$TRUE
}

##########################################################################
#                                                                        #
# function: print_usage                                                  #
#                                                                        #
##########################################################################
#                                                                        #
# prints the usage information                                           #
#                                                                        #
##########################################################################
#                                                                        #
# input:                                                                 #
#                                                                        #
##########################################################################
#                                                                        #
# output:                                                                #
#                                                                        #
##########################################################################

function print_usage() {
cat <<HEREDOC

generate_dependencies.sh
-----------------------------

Usage:

  generate_dependencies.sh DEPENDENCYFILE LIBNAME [ DEPENDENCYFILE1 LIBNAME1 [ DEPENDENCYFILE2 LIBNAME2 [ ... ] ] ] -- SRCFILE1 [ SRCFILE2 [ ... ] ]

Parameters:

  DEPENDENCYFILE   - the dependency file to write
  LIBNAME          - the name of the library, where the .ao, .o
                     files will be found in. (e.g.: libtest )
  DEPENDENCYFILEn  - [ n is a number ]
                     a dependency file to include when checking
                     for dependencies
  LIBNAMEn         - [ n is a number ]
                     the relative path to the library where the 
                     .ao and .o files can be found
                     (e.g.: ../../lib/libcharset )
  SRCFILEn         - [ n is a number ]
                     the files to build the deepndence information
                     for



Description:

This script generates the dependencies for the current directory and stores them in DEPENDENCYFILE. Dependent files are assumed to be in LIBNAME.a for .o files and LIBNAME.al for .ao files. In addition, additional dependency files and their libraries can be supplied in order to allow more sensitive recompilation.
Furthermore only one object may be defined per file. (i.e.: a file "A.as" has
 to declare "A", and nothing else.
 In addition to these restrictions, the resulting output, does not need to be
 minimal. (i.e.: A dependence situation like
    B.ao : C.as
    A.ao : B.ao
  will lead to an output, like
    B.ao : C.as
    A.ao : B.ao C.ao
  if "C" occurs in "A.as".
HEREDOC
}



##########################################################################



if [ $# -lt 3 ]
then
  print_usage
  exit $ERRORARGUMENTS
fi

OUTPUTFILE=$1
SRCLIBNAME=$2
SEPARATORPARAMIDX=3

PARAMETER="echo \${$SEPARATORPARAMIDX}"
while [ `eval $PARAMETER` != "--" -a $SEPARATORPARAMIDX -le $# ]
do
  SEPARATORPARAMIDX=$(($SEPARATORPARAMIDX+1))
  PARAMETER="echo \${$SEPARATORPARAMIDX}"
done
if [ `eval $PARAMETER` != "--" -o $(( $SEPARATORPARAMIDX % 2 )) -ne 1 ]
then
  print_usage
  exit $ERRORARGUMENTS
fi

SCRIPTPATH=`dirname "$0"`

THISIDX=-1


###################################################
#
# retrieving information about the source files in
# the current directory and its subdirectories
#

if [ $# -gt $SEPARATORPARAMIDX ]
then
  for PARAMIDX in `seq $(( $SEPARATORPARAMIDX + 1 )) $#`
  do
    SRCFILE="\${$PARAMIDX}"; SRCFILE=`eval echo $SRCFILE`
    TYPE=`$SCRIPTPATH/get_type_for_file.sh $SRCFILE`
    if [ $? -eq 0 ] 
    then
      THISIDX=${#EXPTYPE[@]}
      ASFILE[$THISIDX]=$SRCFILE
      LIBNAME[$THISIDX]=$SRCLIBNAME
      generate_file_variables
      
      EXPTYPE[$THISIDX]=$TYPE
      LEAFFILE[$THISIDX]=$TRUE
    else
      echo "cannot resolve type for $SRCFILE" >&2
      exit $ERRORUNDEFINEDERROR
    fi
  done
  SRCFILES=$(( $THISIDX + 1 ))
else
  SRCFILES=0
fi



cat >"$OUTPUTFILE" <<HEREDOC
##################################################################
#
# DO NOT EDIT THIS FILE, IT WAS AUTOMATICALLY GENERATED VIA
#
#   $0 $*
#
# in directory: "`pwd`"                                          
#                                                                
##################################################################



##################################################################
#
# DEPENDENCE INFORMATION
#
##################################################################

HEREDOC





###################################################
#
# retrieving information about the source files in
# already built dependency files
#
if [ $SEPARATORPARAMIDX -gt 3 ]
then
  for PARAMIDX in `seq 3 2 $(( $SEPARATORPARAMIDX -1 ))`
  do

    DEPSTATE=0 # 0 waiting for next type
               # 1 waiting for exported type
               # 2 waiting for delimiter
               # 3 waiting for file
    DEPDEPENDENCYFILE="\${$PARAMIDX}"
    DEPDEPENDENCYFILE=`eval echo $DEPDEPENDENCYFILE`
    DEPLIBNAME="\${$(($PARAMIDX + 1 ))}"
    DEPLIBNAME=`eval echo $DEPLIBNAME`

    if [ -f $DEPDEPENDENCYFILE ]
    then
      for GREPPED in `grep \#PROVIDEDTYPE: $DEPDEPENDENCYFILE`
      do
	if [ $DEPSTATE -eq 0 ]
	then
	  DEPSTATE=1
	elif [ $DEPSTATE -eq 1 ]
	then
	  THISIDX=${#EXPTYPE[@]}
	  EXPTYPE[$THISIDX]=$GREPPED
	  DEPSTATE=2
	elif [ $DEPSTATE -eq 2 ]
	then
	  DEPSTATE=3
	elif [ $DEPSTATE -eq 3 ]
	then
	  ASFILE[$THISIDX]=$GREPPED    
	  LIBNAME[$THISIDX]=$DEPLIBNAME
	  generate_file_variables
	  DEPSTATE=0
	else
	  echo "undefined state in aggregating type infromation" >&2
	  exit $ERROEUNDEFINEDERROR
	fi
      done
    else
      echo "could not find dependency file: $DEPDEPENDENCYFILE" >&2
      exit $ERROEUNDEFINEDERROR
    fi
  done
fi
DEPFILES=$(( $THISIDX + 1 - $SRCFILES))

ALLFILES=$(( $THISIDX + 1 ))



###################################################
#
# building dependency information (looking in
# every sources file whether or not one of the 
# exported types occurs)

if [ $SRCFILES -gt 0 ]
then
  for SRCFILEIDX in `seq 0 $(($SRCFILES - 1))`
  do
    SRCFILE=${ASFILE[$SRCFILEIDX]}
    DEPENDENCESTRAO="${AOFILE[$SRCFILEIDX]} : $SRCFILE"
    DEPENDENCESTRO="${OFILE[$SRCFILEIDX]} : ${AOFILE[$SRCFILEIDX]}"
    
    for DEPFILEIDX in `seq 0 $(($ALLFILES - 1))`
    do
      DEPTYPE=${EXPTYPE[$DEPFILEIDX]}
      if [ $DEPFILEIDX -ne $SRCFILEIDX -a ! -z "$DEPTYPE" ]
      then
	if cat "$SRCFILE" | sed -e 's/--.*$//'| sed -e 's/++.*$//'| grep -q -e "\(^\|[^\!\?a-zA-Z0-9]\)$DEPTYPE\([^\!\?a-zA-Z0-9]\|\$\)"
	then
	  # found at least one occurrence of $DEPTYPE
	  # in $SRCFILE. So there is a dependence
#	  DEPENDENCESTRAO="$DEPENDENCESTRAO  ${LIBNAME[$DEPFILEIDX]}.al(${AOFILE[$DEPFILEIDX]})"
#	  DEPENDENCESTRO="$DEPENDENCESTRO  ${LIBNAME[$DEPFILEIDX]}.a(${OFILE[$DEPFILEIDX]})"	
	  DEPENDENCESTRAO="$DEPENDENCESTRAO  ${LIBAOFILE[$DEPFILEIDX]}"
	  DEPENDENCESTRO="$DEPENDENCESTRO  ${LIBOFILE[$DEPFILEIDX]}"	
	  LEAFFILE[$DEPFILEIDX]=$FALSE
	  if [ $DEPFILEIDX -lt $SRCFILES ]
	  then
	    INDEPENDENTFILE[$SRCFILEIDX]=$FALSE
	  fi
	fi
      fi
    done 
  
cat >>"$OUTPUTFILE" <<HEREDOC


#-----------------------------------------------------------------
#PROVIDEDTYPE: ${EXPTYPE[$SRCFILEIDX]} - ${ASFILE[$SRCFILEIDX]}
#-----------------------------------------------------------------
$DEPENDENCESTRAO
$DEPENDENCESTRO
HEREDOC
  done
fi

###################################################

cat >>"$OUTPUTFILE" <<HEREDOC



##################################################################
#
# LEAF FILE INFORMATION
#
##################################################################



LEAFASFILES = \\
HEREDOC
if [ $SRCFILES -gt 0 ]
then
  for SRCFILEIDX in `seq 0 $(($SRCFILES - 1))`
  do
    if [ ${LEAFFILE[$SRCFILEIDX]} -eq $TRUE ]
    then
      echo "	${ASFILE[$SRCFILEIDX]} \\" >>"$OUTPUTFILE"
    fi
  done
fi
cat >>"$OUTPUTFILE" <<HEREDOC

LEAFAOFILES = \\
HEREDOC
if [ $SRCFILES -gt 0 ]
then
  for SRCFILEIDX in `seq 0 $(($SRCFILES - 1))`
  do
    if [ ${LEAFFILE[$SRCFILEIDX]} -eq $TRUE ]
    then
      echo "	${AOFILE[$SRCFILEIDX]} \\" >>"$OUTPUTFILE"
    fi
  done
fi
cat >>"$OUTPUTFILE" <<HEREDOC

LEAFOFILES = \$(LEAFAOFILES:%.ao=%.o)
HEREDOC



###################################################
#
# finding a file to start the library with

if [ $SRCFILES -gt 0 ]
then
  LOWERCASELIBNAMESTUB=`echo -n "$LIBNAME" | sed -e 's/^lib//'| awk '{printf("%s",tolower($0))}'`
  for SRCFILEIDX in `seq 0 $(($SRCFILES - 1))`
  do
    SRCFILE=${ASFILE[$SRCFILEIDX]}
    if [ ${INDEPENDENTFILE[$SRCFILEIDX]} -eq $TRUE ]
    then
      if [ `grep -c "^\#include[\ \t]*\"[\ \t]*$LOWERCASELIBNAMESTUB\(\.as\)\?[\ \t]*\"[\ \t]*" "$SRCFILE"` -eq 0 ]
      then
	STARTIDX=$SRCFILEIDX
	
	#a leaf file, that can start the library is the best one can find
	if [ ${LEAFFILE[$SRCFILEIDX]} -eq $TRUE ]
	then
	  LEAFSTARTIDX=$SRCFILEIDX
	fi
      fi
    fi 
  done

  if [ ! -z "$LEAFSTARTIDX" ]
  then
    STARTIDX=$LEAFSTARTIDX
  fi

  if [ -z "$STARTIDX" ]
  then
    echo "cannot find file to start the library with" >&2
    exit $ERRORUNDEFINEDERROR
  fi

  cat >>"$OUTPUTFILE" <<HEREDOC
  
  
#-----------------------------------------------------------------
# FILES TO BUILD FOR THE LIBRARY
#-----------------------------------------------------------------
BUILDASFILES = \\
	${ASFILE[$STARTIDX]} \\
HEREDOC
  for SRCFILEIDX in `seq 0 $(($SRCFILES - 1))`
  do
    if [ ${LEAFFILE[$SRCFILEIDX]} -eq $TRUE -a $SRCFILEIDX -ne $STARTIDX ]
    then
      echo "	${ASFILE[$SRCFILEIDX]} \\" >>"$OUTPUTFILE"
    fi
  done
  cat >>"$OUTPUTFILE" <<HEREDOC

BUILDAOFILES = \\
	${AOFILE[$STARTIDX]} \\
HEREDOC
  for SRCFILEIDX in `seq 0 $(($SRCFILES - 1))`
  do
    if [ ${LEAFFILE[$SRCFILEIDX]} -eq $TRUE -a $SRCFILEIDX -ne $STARTIDX ]
    then
      echo "	${AOFILE[$SRCFILEIDX]} \\" >>"$OUTPUTFILE"
    fi
  done
  cat >>"$OUTPUTFILE" <<HEREDOC

BUILDOFILES = \$(BUILDAOFILES:%.ao=%.o)
HEREDOC

fi



