#!/bin/sh

##########################################################################
#                                                                        #
#                 generate_documentation_aldordoc.sh                     #
#                                                                        #
##########################################################################
#                                                                        #
# A script for extracting the documentation of an Aldor source file and  #
# writing the documentation in aldordoc format.                          #
#                                                                        #
##########################################################################
#                                                                        #
# Copyright (C) 2005  Research Institute for Symbolic Computation, J.    #
# Kepler University, Linz, Austria                                       #
#                                                                        #
# Written by Christian Aistleitner                                       #
#                                                                        #
# This program is free software; you can redistribute it and/or          #
# modify it under the terms of the GNU General Public License version 2, #
# as published by the Free Software Foundation.                          #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program; if not, write to the Free Software            #
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,             #
# MA  02110-1301, USA.                                                   #
#                                                                        #
##########################################################################

ERROROK=0
ERRORUNDEFINEDERROR=100
ERRORARGUMENTS=$(($ERRORUNDEFINEDERROR+1))

MAXADSNIPPETWIDTH=50

function print_usage() {
cat <<HEREDOC

generate_documentation_aldordoc.sh
-----------------------------

Usage:

  generate_documentation_aldordoc.sh ENVIRONMENTMAKEFILE ALDORFILE TEXFILE 

Parameters:

  ENVIRONMENTMAKEFILE - a Makefile defining properties for the aldor files
                        (e.g.: environment.mk)
  ALDORFILE   - An Aldor source cade file (.as, or .ai)
  TEXFILE     - The file to write the documentation of ALDORFILE to


Description:

This script generates an aldordoc .tex file for ALDORFILE

HEREDOC
}

if [ $# -ne 3 ]
then
  print_usage
  exit $ERRORARGUMENTS
fi


ENVIRONMENTMK=$1
ALDORFILE=$2
TEXFILE=$3
if [ -e "$TEXFILE" ] 
then
  rm "$TEXFILE"
fi
exec 1<>$TEXFILE
## arguments are ok #################################################
SCRIPTPATH=`dirname "$0"`
EXPORTEDTYPE=`$SCRIPTPATH/get_type_for_file.sh $ALDORFILE`
DOCDIR=`$SCRIPTPATH/query_environment.sh DOCDIR "$ENVIRONMENTMK"`
DOCLIBRARYDOCDIR=`$SCRIPTPATH/query_environment.sh DOCLIBRARYDOCDIR "$ENVIRONMENTMK"`
LIBDIR=`$SCRIPTPATH/query_environment.sh LIBDIR "$ENVIRONMENTMK"`
SRCDIR=`$SCRIPTPATH/query_environment.sh SRCDIR "$ENVIRONMENTMK"`
PROJECTROOT=`$SCRIPTPATH/query_environment.sh PROJECTROOT "$ENVIRONMENTMK"`
DEPENDENCYFILE=`$SCRIPTPATH/query_environment.sh DEPENDENCYFILE "$ENVIRONMENTMK"`
NOIDCHAR="[^0-9a-zA-Z\!\?]"
NOIDCHAROREOL="\($NOTIDCHOR\|$\)"
TRUE=1
FALSE=0

#counting the number of functions already identified
FUNCS=0 #stored elements in FUNCS FUNCSPARAM FUNCSCONDITION FUNCSDOC
TYPES=0 #stored elements in TYPES TYPESPARAM TYPESCONDITION TYPESDOC
MACROS=0 #stored elements in MACRO

#the following variable is used to work properly on the last line 
#of a file, as this line contains no final CR and therefor "read"
#already indicates a EOF.
LASTREADGAVEEOF=$FALSE

#if $FOUNDDEFAULT is $TRUE, the default block of a with statement
#is reached. Therefore no further functions are to be considered 
#by convention.
FOUNDDEFAULT=$FALSE


#breaks lines and adds tabbing in adsnippets environments
function print_snippet_line() {
echo "$SNIPPETTABBING$SNIPPETLINE"
#  while [ ! -z "$SNIPPETLINE" ]
#  do
#    SNIPPETHEAD="`echo $SNIPPETLINE | sed -e 's/^\(.\{60\}\)\(.*\)$/\1/'`"
#    SNIPPETTAIL="`echo $SNIPPETLINE | sed -e 's/^\(.\{60\}\)\(.*\)$/\2/'`"
#    echo -n "$SNIPPETTABBING"
#    if [ "$SNIPPETLINE" = "$SNIPPETTAIL" ]
#    then
#      echo $SNIPPETTAIL
#      SNIPPETLINE=""
#    else
#      echo $SNIPPETHEAD
#      SNIPPETLINE=$SNIPPETTAIL
#    fi
#  done  
}

#sets LINE
function trim_line() {
  LINE="`echo $LINE | sed -e 's/--.*$//' -e 's/^[\ \t]*\([^\ \t]\)/\1/'`"
}

#sets LASTREADGAVEEOF
function check_eof() {
  if [ $? -ne 0 ]
  then
    if [ $LASTREADGAVEEOF -eq $TRUE ]
    then
      ERROR="premature end of input"
      error
    else
      LASTREADGAVEEOF=$TRUE
    fi
  fi 
}

#sets LINE
function myread() {
  read -r LINE
  check_eof
  trim_line
  while [ "`echo $LINE | grep -c -e \"^[\ \t]*$\|^#\"`" = "1" ]
  do
    read -r LINE
    check_eof
    trim_line
  done
}

#sets LINE
function grab_documentation() {
  while [ "`echo $LINE | grep -c -e \"^[\ \t]*+++\"`" = "1" ] 
  do
    DOCUMENTATION="$DOCUMENTATION \n `echo $LINE | sed -e 's/.*+++//' -e 's/\\\/\\\\\\\/g'`"
    myread
  done
}

#sets LINE
function read_to_next_expr() {
  while [ "`echo $LINE | grep -c -e \"^[\ \t]*$\"`" = "1" ]
  do
    myread
  done
  LINE="`echo $LINE | sed -e 's/^[\ \t]*\([^\ \t].*\)$/\1/'`"
}

#needs $ERROR, $ALDORFILE, and $EXPORTEDTYPE
function error() {
  echo "problems parsing file $ALDORFILE for documentation of $EXPORTEDTYPE"  >&2
  echo "$ERROR"  >&2
  exit 1
}

#sets LINE and READSTATUS
function parse_function() {
  NAME=
  PARAMETER=
  CONDITION=
  DOCUMENTATION=

  if [ "`echo $LINE | grep -c -e \"^if[^a-zA-Z0-9\!\?].*$\"`" = "1" ]
  then
    if [ "`echo $LINE | grep -c -e \"then\"`" = "0" ]
    then
      ERROR="the line of the \"if\" does not contain a \"then\""
      error
    fi

    CONDITION="`echo $LINE | sed -e \"s/^\(.*${NOIDCHAR}then\).*$/\1/\"`";
    LINE="`echo $LINE | sed -e \"s/^.*${NOIDCHAR}then\(.*\)$/\1/\"`";

    read_to_next_expr

    #$LINE has to start with "{"
    if [ "`echo $LINE | grep -c -e \"^{\"`" = "0" ]
    then
      ERROR="the line following the \"if\" does not start in \"{\""
      error
    fi   

    LINE="`echo $LINE | sed -e 's/^{//'`";
    read_to_next_expr
  else
    CONDITION=""
  fi

  #fail when using postdoc instead of predoc
  if [ "`echo $LINE | grep -c -e \"^++[^+]\"`" = "1" ]
  then
    ERROR="encountered post-documentiation (++). But this script can only deal with pre-documentation (+++)"
    error    
  fi

  #fail when nesting with's
  if [ "`echo $LINE | grep -c -e \"^with[\{\(\ \t]\"`" = "1" ]
  then
    ERROR="encountered a \"with\" clause, where a function name or pre-doc should be. Remember that it is not allowed to nest with statements."
    error    
  fi

  #$LINE contains either predoc or the function to parse
  #if the function is in an if clause, this if clause is in $CONDITION
  grab_documentation

  #$LINE contains the function to parse
  #DOCUMENTATION contains the functions documentation
  if [ "`echo $LINE | grep -c -e \"with\"`" = "1" ]
  then
    NAME="`echo $LINE | sed -e 's/^\(.*\)with.*$/\1/'`";
    LINE="`echo $LINE | sed -e 's/^.*\(with.*\)$/\1/'`";
  else
    NAME=$LINE
    myread
  fi

  #$LINE starts with "}" for conditional expressions,
  #otherwise the next character after the current function
  #$NAME contains the function name, parameters and possibly a trailing ";"
  NAME="`echo $NAME | sed -e 's/;$//'`";

  #$NAME contains the function name, parameters and no trailing ";"
  PARAMETER="`echo $NAME | sed -e \"s/^[a-z<>\+\-\*\/\\][0-9a-zA-Z\!\?]*[\ \t]*\(${NOIDCHAROREOL}*\)/\1/\"`";
  NAME="`echo $NAME | sed -e \"s/^\([a-z<>\+\-\*\/\\][0-9a-zA-Z\!\?]*\)$NOIDCHAROREOL.*$/\1/\"`";

  #$NAME contains the function name, no parameters and no trailing ";"
  #$PARAMETER contains the parameters (including a leading :)
  if [ "`echo $PARAMETER | grep -c -e \"^:\"`" = "1" ]
  then
    ISDOMAIN=$FALSE
    PARAMETER="`echo $PARAMETER | sed -e 's/^:[\ \t]*\([^\ \t]\)/\1/'`";
  else
    ISDOMAIN=$TRUE
  fi

  if [ ! -z "$CONDITION" ]
  then
    if [ "`echo $LINE | grep -c -e \"^}\"`" = "0" ]
    then
      ERROR="could not find trailing \"}\" for the \"if\" statement -- remember that you are not allowed to nest ifs and only one definition per if is allowed."
      error
    fi   
    LINE="`echo $LINE | sed -e 's/^}//'`";
    read_to_next_expr
  fi   

  if [ "$NAME" = "default" ]
  then
    FOUNDDEFAULT=$TRUE
  else
    if [ $ISDOMAIN -eq $FALSE ]
    then
      FUNCSNAME[$FUNCS]=$NAME
      FUNCSPARAMETER[$FUNCS]=$PARAMETER
      FUNCSCONDITION[$FUNCS]=$CONDITION
      FUNCSDOCUMENTATION[$FUNCS]=$DOCUMENTATION
      FUNCS=$(($FUNCS+1))
    else
      TYPESNAME[$TYPES]=$NAME
      TYPESPARAMETER[$TYPES]=$PARAMETER
      TYPESCONDITION[$TYPES]=$CONDITION
      TYPESDOCUMENTATION[$TYPES]=$DOCUMENTATION
      TYPES=$(($TYPES+1))
    fi
  fi
  
}

#prints an aldor expression with parethesis
#the whole expression has to be stored in $LINE
#CURCOL has to be set to the number of characters on the current line
#IDENTATION has to be set to a string used for identation
#IDENTATIONCOLS has to denote the lentght of IDENTATION
function print_aldor_expression() {
  trim_line
  PARSEDEXPRESSION=""
  if [ -z "$LINE" ]
  then
      ERROR="trying to parse an aldor expression from an empty string"
      error
  fi
  PARLEVEL=0
  LASTCHAR=" "
  FIRSTCHAR="`echo x$LINE | sed -e 's/^x\(.\).*$/\1/'`"
  # 0 = empty; 1 = one character; 2 = undetermined length
  FIRSTCHARTYPE=1
  while [ "`echo $FIRSTCHAR | grep -c -e '[,;)}]'`" = "0" -o "$PARLEVEL" != "0" ]
  do
#    echo $CURCOL >&2
    case "$FIRSTCHAR" in
      ";")
	   PARSEDEXPRESSION="$PARSEDEXPRESSION$FIRSTCHAR\\n$IDENTATION"
	   CURCOL=$IDENTATIONCOLS
	   FIRSTCHAR=""
	   FIRSTCHARTYPE=0
	   ;;
      "{")
	   IDENTATION="${IDENTATION}  "
	   IDENTATIONCOLS=$((IDENTATIONCOLS+2))
	   PARLEVEL=$((PARLEVEL+1))
	   PARSEDEXPRESSION="$PARSEDEXPRESSION$FIRSTCHAR\\n$IDENTATION"
	   CURCOL=$IDENTATIONCOLS
	   FIRSTCHAR=""
	   FIRSTCHARTYPE=0
	   ;;
      "(")
	   IDENTATION="${IDENTATION}  "
	   IDENTATIONCOLS=$((IDENTATIONCOLS+2))
	   PARLEVEL=$((PARLEVEL+1))
	   FIRSTCHAR="( "
	   FIRSTCHARTYPE=2
	   ;;
      ")")
	   IDENTATION="`echo \"x$IDENTATION\" | sed -e 's/x\(.*\)..$/\1/'`"
	   IDENTATIONCOLS=$((IDENTATIONCOLS-2))
	   PARLEVEL=$((PARLEVEL-1))
	   ;;
      "}")
	   IDENTATION="`echo \"x$IDENTATION\" | sed -e 's/x\(.*\)..$/\1/'`"
	   IDENTATIONCOLS=$((IDENTATIONCOLS-2))
	   PARLEVEL=$((PARLEVEL-1))
	   PARSEDEXPRESSION="`echo \"$PARSEDEXPRESSION\" | sed -e 's/\([^ \t]\)\([ \t]\|\\\n\)*$/\1/'`"
	   PARSEDEXPRESSION="$PARSEDEXPRESSION\n$IDENTATION$FIRSTCHAR"      
	   CURCOL=$((IDENTATIONCOLS+1))
	   FIRSTCHAR=""
	   FIRSTCHARTYPE=0
	   ;;
      "\\")
	    FIRSTCHAR="\\\\"
	    ;;
    esac
#    if [ "`echo $FIRSTCHAR | grep -c -e '[;]'`" = "1" ]
#    then
#      PARSEDEXPRESSION="$PARSEDEXPRESSION$FIRSTCHAR\\n$IDENTATION"
#      FIRSTCHAR=""
#    elif [ "`echo $FIRSTCHAR | grep -c -e '[{]'`" = "1" ]
#    then
#      IDENTATION="${IDENTATION}  "
#      PARLEVEL=$((PARLEVEL+1))
#      PARSEDEXPRESSION="$PARSEDEXPRESSION$FIRSTCHAR\\n$IDENTATION"
#      FIRSTCHAR=""#    elif [ "`echo $FIRSTCHAR | grep -c -e '[(]'`" = "1" ]
#    then
#      FIRSTCHAR="( "
#      PARLEVEL=$((PARLEVEL+1))
#    elif [ "`echo $FIRSTCHAR | grep -c -e '[)]'`" = "1" ]
#    then
#      PARLEVEL=$((PARLEVEL-1))
#    elif [ "`echo $FIRSTCHAR | grep -c -e '[}]'`" = "1" ]
#    then
#      IDENTATION="`echo \"x$IDENTATION\" | sed -e 's/x\(.*\)..$/\1/'`"
#      PARLEVEL=$((PARLEVEL-1))
#      PARSEDEXPRESSION="`echo \"$PARSEDEXPRESSION\" | sed -e 's/\([^ \t]\)\([ \t]\|\\\n\)*$/\1/'`"
#      PARSEDEXPRESSION="$PARSEDEXPRESSION\n$IDENTATION$FIRSTCHAR"      
#      FIRSTCHAR=""
#    elif [ "`echo $FIRSTCHAR | grep -c -e '[;]'`" = "1" ]
#    then
#      PARLEVEL=$((PARLEVEL-1))
#    elif [ "`echo $FIRSTCHAR | grep -c -e '[\\]'`" = "1" ]
#    then
#      FIRSTCHAR="\\\\"
#    fi
#    if [ `echo -e "$PARSEDEXPRESSION" | tail -n 1 | wc --chars` -gt $MAXADSNIPPETWIDTH ]
    PARSEDEXPRESSION="$PARSEDEXPRESSION$FIRSTCHAR"
    if [ $FIRSTCHARTYPE -eq 1 ]
    then
      CURCOL=$((CURCOL+1))
    elif [ $FIRSTCHARTYPE -eq 2 ]
    then
      CURCOL=$((CURCOL+`echo -e "x${FIRSTCHAR}x" | tail -n 1 | wc --chars`-3))
    fi
    if [ $CURCOL -ge $MAXADSNIPPETWIDTH ]
    then
      CURRENTLINERIGHT=""
      if [ "`echo x${FIRSTCHAR}x | grep -c -e '^x[0-9a-zA-Z\!\?]x$'`" = "1" ]
      then
	# the last written character is a letter. So we are in the middle of a word
	CURRENTLINE="`echo \"$PARSEDEXPRESSION\" | sed -e 's/^.*\\\n\(\([^\\]\|\\[^n]\)*\)$/\1/'`"
	CURRENTLINERIGHT="`echo \"$CURRENTLINE\" | sed -e 's/^.*[^0-9a-zA-Z\!\?]\([0-9a-zA-Z\!\?]*\)$/\1/'`"
	if [ `echo -e "$CURRENTLINERIGHT" | tail -n 1 | wc --chars` -lt 10 ]
	then
	  # last word has few characters on the current line. So move it to the next line

	  CURRENTLINELEFT="`echo \"$CURRENTLINE\" | sed -e 's/^\(.*[^0-9a-zA-Z\!\?]\)[0-9a-zA-Z\!\?]*$/\1/'`"
	  if [ "$CURRENTLINELEFT$CURRENTLINERIGHT" = "$PARSEDEXPRESSION" ]
	  then
	    #still in first line. all the information is in CURRENTLINELEFT and CURRENTLINERIGHT
	    PARSEDEXPRESSION=""
	  else
	    #more than one line ... cut off last line
	    PARSEDEXPRESSION="`echo \"$PARSEDEXPRESSION\" | sed -e 's/^\(.*\\\n\)\([^\\]\|\\[^n]\)*$/\1/'`"
	  fi
	  PARSEDEXPRESSION="$PARSEDEXPRESSION${CURRENTLINELEFT}\\n$IDENTATION  $CURRENTLINERIGHT"
	  CURCOL=$((IDENTATIONCOLS+2+`echo x${CURRENTLINERIGHT}x | wc --chars`-3))
	else
	  # last word has lots characters on the current line. Trying to break at capital letter	  

	  #sed is picky about character classes in the following statements. Therefore, they are written down explicitely :(
	  #CURRENTLINERIGHT="`echo \"$CURRENTLINE\" | sed -e 's/^.*[^A-Z]\([A-Z][0-9a-zA-Z\!\?]*\)$/\1/'`"
	  CURRENTLINERIGHT="`echo \"$CURRENTLINE\" | sed -e 's/^.*[^ABCDEFGHIJKLMNOPQRSTUVWXYZ]\([ABCDEFGHIJKLMNOPQRSTUVWXYZ][0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ\!\?]*\)$/\1/'`"
	  if [ "$CURRENTLINE" != "$CURRENTLINERIGHT" ]
	  then
	    #the last word contains a capital letter
	    if [ `echo -e "$CURRENTLINERIGHT" | tail -n 1 | wc --chars` -lt 10 ]
	    then
	      # last word has a capital letter not too far of the end. Cut at capital letter.
	      
	      #sed is picky about character classes in the following statements. Therefore, they are written down explicitely :(
	      #CURRENTLINELEFT="`echo \"$CURRENTLINE\" | sed -e 's/^\(.*[^A-Z]\)[A-Z][0-9-a-zA-Z\!\?]*$/\1/'`"
	      CURRENTLINELEFT="`echo \"$CURRENTLINE\" | sed -e 's/^\(.*[^ABCDEFGHIJKLMNOPQRSTUVWXYZ]\)[ABCDEFGHIJKLMNOPQRSTUVWXYZ][0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ\!\?]*$/\1/'`"
	      if [ "$CURRENTLINELEFT$CURRENTLINERIGHT" = "$PARSEDEXPRESSION" ]
	      then
		#still in first line. all the information is in CURRENTLINELEFT and CURRENTLINERIGHT
		PARSEDEXPRESSION=""
	      else
		#more than one line ... cut off last line
		PARSEDEXPRESSION="`echo \"$PARSEDEXPRESSION\" | sed -e 's/^\(.*\\\n\)\([^\\]\|\\[^n]\)*$/\1/'`"
	      fi
	      PARSEDEXPRESSION="$PARSEDEXPRESSION${CURRENTLINELEFT}\\n$IDENTATION  $CURRENTLINERIGHT"
	      CURCOL=$((IDENTATIONCOLS+2+`echo x${CURRENTLINERIGHT}x | wc --chars`-3))
	    fi
	  fi	  
	fi
      fi
      if [ "`echo x${FIRSTCHAR}x | grep -c -e '^x-x$'`" = "1" ]
      then
	if [ "`echo x${LINE}x | grep -c -e '^x->.*'`" = "1" ]
	then
	  #breaking an arrow "->"
	  PARSEDEXPRESSION="`echo \"$PARSEDEXPRESSION\" | sed -e 's/^\(.*\).$/\1/'`"
	  PARSEDEXPRESSION="$PARSEDEXPRESSION\\n$IDENTATION  -"
	  CURCOL=$((IDENTATIONCOLS+2))
	fi
      fi
      if [ $CURCOL -ge $MAXADSNIPPETWIDTH ]
      then
	#nane of the heuristics could break the line. So we perform a hard break
	PARSEDEXPRESSION="$PARSEDEXPRESSION\\n$IDENTATION  "
	CURCOL=$((IDENTATIONCOLS+2))
      fi
    fi
    LINE="`echo x$LINE | sed -e 's/^x.//'`"
#echo $LINE >&2
    if [ -z "$LINE" ]
    then
      if [ "$PARLEVEL" = "0" ]
      then
	#proper ending, bailing out with ";"
	FIRSTCHAR=";"
      else
	#some characters missing
	ERROR="ran out of characters befor aldor expression ended"
	error
      fi
    else
      LASTCHAR="$FIRSTCHAR"
      FIRSTCHAR="`echo x$LINE | sed -e 's/^x\(.\).*$/\1/'`"
      FIRSTCHARTYPE=1
    fi
    LONGFIRSTCHAR=0
  done
  echo -e "$PARSEDEXPRESSION"
}

#breaking lines for declaration parameters
function format_declaration_parameters() {
  if [ "`echo $DECLARATIONPARAMETERS | grep -c -e '[^(]*(' `" != "1"  -o "`echo $DECLARATIONPARAMETERS | grep -c -e ')[\ ]*$' `" != "1" -o -z "$DECLARATIONPARAMETERS" ]
  then
      ERROR="the parameters of the category or domain are either null or do not contain starting or ending parenthesis"
      error
  fi
  LINE="`echo $DECLARATIONPARAMETERS | sed -e 's/[^(]*(//' -e 's/)[\ ]*$//' `"
  IDENTLEVEL=1
  trim_line
  while [ ! -z "$LINE" ]
  do
    PARAMETERNAME="`echo $LINE | sed -e 's/^\([^:]*\):.*$/\1/'`";
    LINE="`echo $LINE | sed -e 's/^\([^:]*\)://'`";
    trim_line
    echo -n $PARAMETERNAME
    PARAMETERNAMELEN="`echo $PARAMETERNAME | wc --chars`"
    if [ $PARAMETERNAMELEN -lt 5 ]
    then
      while [ $PARAMETERNAMELEN -lt 5 ]
      do
	echo -n " "
	PARAMETERNAMELEN=$((PARAMETERNAMELEN+1))
	CURCOL=5
      done    
    else
      CURCOL=$PARAMETERNAMELEN
    fi
    echo -n ": "
    CURCOL=$((CURCOL+2-1))
    IDENTATION="      "
    IDENTATIONCOLS=6
    print_aldor_expression
    FIRSTCHAR="`echo $LINE | sed -e 's/\(.\).*$/\1/'`"
    if [ "`echo $FIRSTCHAR | grep -c -e '[);]'`" = "1" ]
    then
      LINE=""
    elif [ "$FIRSTCHAR" = "," ]
    then
      LINE="`echo $LINE | sed -e 's/^.//'`"    
      trim_line
    fi
  done
}

#taking $ALDORFILE as input
exec <$ALDORFILE

#SEEKINGDECLARATION is 0 as long as we are looking for the declaration of $EXPORTEDTYPE
SEEKINGDECLARATION=0

#READSTATUS is 0 as long as we are not at eaof of $ALDORFILE
READSTATUS=0

myread
#filling DOCUMENTATION with the documentation for $EXPORTEDTYPE and looking for $EXPORTEDTYPE's declaration
while [ $SEEKINGDECLARATION -eq 0 ]
do
      grab_documentation
      ISDECLARATION=`echo $LINE | grep -c -e "^\(define[\ \t]*\|extend[\ \t]*\)\?$EXPORTEDTYPE$NOIDCHAROREOL"`
      if [ "$ISDECLARATION" = "1" ] 
      then
	SEEKINGDECLARATION=1
      else
	ISMACROLINE=`echo $LINE | grep -c -e "^macro[\ \t]"`
	if [ "`echo $LINE | grep -c -e '^macro[\ \t]'`" = "1" ] 
	then
	  MACRO[$MACROS]="`echo $LINE | sed -e 's/^macro[\ \t]//'`"
	  MACROS=$(($MACROS+1))
	else
	  MACROS=0
	fi
	DOCUMENTATION=""
	myread
      fi
done

DECLARATIONDOCUMENTATION=$DOCUMENTATION
DOCUMENTATION=""

# $LINE holds the first line of the definiton of $EXPORTEDTYPE
# $TYPEDOC holds the documentation for $EXPORTEDTYPE
# $DECLARATION is to be filled with the whole declaration
DECLARATION=`echo $LINE | sed -e "s/^\(\(define[\ \t]*\|extend[\ \t]*\)\?$EXPORTEDTYPE\)[^0-9a-zA-Z\!\?].*$/\1/"`
DECLARATIONTYPE=`echo $LINE | sed -e "s/^\(\(define\|extend*\)\?[\ \t]*$EXPORTEDTYPE\)[^0-9a-zA-Z\!\?].*$/\2/"`
LINE=`echo $LINE | sed -e "s/^\(\(define[\ \t]*\|extend[\ \t]*\)\?$EXPORTEDTYPE\)\($NOIDCHAROREOL.*\)$/\3/"`

read_to_next_expr
# $LINE should either start with ":" if $EXPORTEDTYPE has no parameters
# or "(" if $EXPORTEDTYPE takes parameters

#parsing the parameters for the $EXPORTEDTYPE
DECLARATIONPARAMETERS=
if [ "`echo $LINE | grep -c -e '^('`" == "1" ]
then
  LINE=`echo $LINE | sed -e "s/^(//"`
  DECLARATIONPARAMETERS="("
  OPENPARENCOUNT=1
  while [ $OPENPARENCOUNT -gt 0 ]
  do
    #add $LINE as long as there are no parenthesis in it
    while [ "`echo $LINE | grep -c -e '^[^()]*$'`" == "1" ]
    do
      DECLARATIONPARAMETERS="$DECLARATIONPARAMETERS$LINE"
      myread
    done

    #LINE has a parenthesis somewhere
    DECLARATIONPARAMETERS="$DECLARATIONPARAMETERS`echo $LINE | sed -e 's/^\([^()]*\)[()].*$/\1/'`"
    LINE="`echo $LINE | sed -e 's/^[^()]*\([()].*\)$/\1/'`"

    #LINE has a parenthesis at first position
    if [ "`echo $LINE | grep -c -e '^('`" == "1" ]
    then
      #parenthesis is opening parenthesis
      OPENPARENCOUNT=$(($OPENPARENCOUNT+1))
      DECLARATIONPARAMETERS="$DECLARATIONPARAMETERS("
      LINE="`echo $LINE | sed -e 's/^(//'`"
    elif [ "`echo $LINE | grep -c -e '^)'`" == "1" ]
    then
      #parenthesis is closing parenthesis
      OPENPARENCOUNT=$(($OPENPARENCOUNT-1))
      DECLARATIONPARAMETERS="$DECLARATIONPARAMETERS)"
      LINE="`echo $LINE | sed -e 's/^)//'`"
    fi
  done
fi

read_to_next_expr
# $LINE should with the ":" for $EXPORTEDTYPE's type
if [ "`echo $LINE | grep -c -e '^:'`" != "1" ]
then
  ERROR="could not find the \":\" before $EXPORTEDTYPE's type"
  error
fi

#skipping ":"
LINE="`echo $LINE | sed -e 's/^:[\ \t]*\([^\ \t].*\)$/\1/'`"
read_to_next_expr

#$LINE holds the type of $EXPORTEDTYPE
if [ "`echo $LINE | grep -c -e \"^Category$NOIDCHAROREOL\"`" == "1" ]
then
  #$EXPORTEDTYPE is a category
  EXPORTEDTYPEISCATEGORY=1
  LINE=`echo $LINE | sed -e "s/^Category//"`
  read_to_next_expr
  if [ "`echo $LINE | grep -c -e \"^==\"`" != "1" ]
  then
    ERROR="could not find the \"==\" after the type \"Category\""
    error
  fi
  LINE=`echo $LINE | sed -e "s/^==//"`
  read_to_next_expr
else
  #$EXPORTEDTYPE is a domain
  EXPORTEDTYPEISCATEGORY=0
fi

#$LINE holds the beginning of with part. This is either 
#  "A with { B }",
#  "A ...", or
#  "with { B }"
if [ "`echo $LINE | grep -c -e \"^with$NOIDCHAR\"`" != "1" ]
then
  #if $EXPORTEDTYPE is of the form
  #  $EXPORTEDTYPE: A == ...
  #then an additional with {} is introduced right away
  if [ "`echo $LINE | grep -c -e \"==[^)]*$\"`" == "1" ]
  then
    # "with {}" part is missing, so adding it
    LINE="`echo $LINE | sed -e \"s/\(==[^)]*\)$/with { } \1/"`"
  fi

  #parsing A
  parse_function

  if [ "`echo $LINE | grep -c -e \"^\(==\|;\|with\)\"`" != "1" ]
  then
    ERROR="no proper ending for the \"with\" part found"
    error
  fi

  if [ "`echo $LINE | grep -c -e \"^\(==\|;\)\"`" == "1" ]
  then
    # "with {}" part is missing, so adding it
    LINE="with { } $LINE"
  fi
fi

#$LINE starts by "with"
if [ "`echo $LINE | grep -c -e \"^with\"`" != "1" ]
then
  ERROR="could not find \"with\" part"
  error
fi
LINE="`echo $LINE | sed -e \"s/^with//"`"
read_to_next_expr

#$LINE starts by the "{" after the with statement
if [ "`echo $LINE | grep -c -e \"^{\"`" != "1" ]
then
  ERROR="could not find \"{\" after the with part"
  error
fi
LINE="`echo $LINE | sed -e \"s/^{//"`"
read_to_next_expr

#$LINE holds either a function, predoc for a function, or "}"
while [ "`echo $LINE | grep -c -e \"^}\"`" != "1" -a $FOUNDDEFAULT -eq $FALSE ]
do
  #parsing the function
  parse_function
done

echo "\clearpage"
if [ "$DECLARATIONTYPE" = "extend" ]
then
  echo "\subsection{$EXPORTEDTYPE (extension)}"
else
  echo "\subsection{$EXPORTEDTYPE}"
fi
echo "\addefinetype{$EXPORTEDTYPE}"
if [ $MACROS -ne 0 ]
then
  echo "\begin{adlocalmacros}"
  echo "\begin{adsnippet}"
  for IDX in `seq 0 $(($MACROS-1))`
  do
    echo "${MACRO[$IDX]}"
  done
  echo " \end{adsnippet}"
  echo "\end{adlocalmacros}"
fi
if [ ! -z "$DECLARATIONPARAMETERS" ]
then
  echo "\begin{adparameters}"
  echo " \begin{adsnippet}"
  #echo "  $DECLARATIONPARAMETERS"
  format_declaration_parameters
  echo " \end{adsnippet}"
  echo "\end{adparameters}"
fi
echo -e $DECLARATIONDOCUMENTATION 

#basic description of exported types and names
HASEXPORTSFROM=0
if [ $TYPES -gt 0 ] 
then
  echo "\begin{adexportedcategories}"
  for IDX in `seq 0 $(($TYPES-1))`
  do
    if [ "${TYPESNAME[$IDX]}" = "export" ]
    then
      #these exports are printed AFTER the functions and constants
      HASEXPORTSFROM=1      
    else
      if [ -z "${TYPESCONDITION[$IDX]}" ]
      then
	echo " \item \adtype{${TYPESNAME[$IDX]}}\adcode|${TYPESPARAMETER[$IDX]}|"
      else
	echo " \item \adcode|${TYPESCONDITION[$IDX]} {|"
	echo " \item \hspace{1em}\adtype{${TYPESNAME[$IDX]}}\adcode|${TYPESPARAMETER[$IDX]}|"
	echo " \item \adcode|}|"
      fi
    fi
  done
  echo "\end{adexportedcategories}"
fi
if [ $FUNCS -gt 0 ] 
then
  echo "\begin{adexportednames}"
  for IDX in `seq 0 $(($FUNCS-1))`
  do
    #signature
    if [ -z "${FUNCSCONDITION[$IDX]}" ]
    then
      echo " \item \adname{${FUNCSNAME[$IDX]}: ${FUNCSPARAMETER[$IDX]}}\adcode|: ${FUNCSPARAMETER[$IDX]}|"
    else
      echo " \item \adcode|${FUNCSCONDITION[$IDX]} {|"
      echo " \item \hspace{1em}\adname{${FUNCSNAME[$IDX]}: ${FUNCSPARAMETER[$IDX]}}\adcode|: ${FUNCSPARAMETER[$IDX]}|"
      echo " \item \adcode|}|"
    fi
    #short doc
    if [ `echo ${FUNCSDOCUMENTATION[$IDX]} | grep -c "\\\\\begin{addescription}"` -gt 0 ]
    then
      if [ `echo ${FUNCSDOCUMENTATION[$IDX]} | grep -c "\\\\\end{addescription}"` -gt 0 ]
      then	
	SHORTDOC="`echo ${FUNCSDOCUMENTATION[$IDX]} | sed -e s/^.*begin{addescription}/\\\\\\\\\\\\\\\hspace{2em}\\\\\\\\\\\\\\\begin{adshortdescription}/ -e s/end{addescription}.*$/end{adshortdescription}/`"
	echo -e \\item $SHORTDOC
      fi
    fi
  done
  echo "\end{adexportednames}"
fi

#printing the "export from" of the with statement
if [ $HASEXPORTSFROM -ne 0 ]
then
  echo "\begin{adexportsfrom}"
  for IDX in `seq 0 $(($TYPES-1))`
  do
    if [ "${TYPESNAME[$IDX]}" = "export" ]
    then
      if [ -z "${TYPESCONDITION[$IDX]}" ]
      then
	echo " \item \adcode|`echo ${TYPESPARAMETER[$IDX]}| sed -e s/from[\ \t]*//`|"
      else
	echo " \item \adcode|${TYPESCONDITION[$IDX]} {|"
	echo " \item \hspace{1em}\adcode|`echo ${TYPESPARAMETER[$IDX]}| sed -e s/from[\ \t]*//`|"
	echo " \item \adcode|}|"
      fi
    fi
  done
  echo "\end{adexportsfrom}"
fi

IMPLFILE="`echo $TEXFILE | sed -e 's/.tex$/.exp/'`"
echo "\IfFileExists{$IMPLFILE}{\input{$IMPLFILE}}{}"

#more detailed description of exported types and names
if [ $FUNCS -gt 0 ] 
then
  for IDX in `seq 0 $(($FUNCS-1))`
  do
    echo "\clearpage"
    ESCAPEDFUNCNAME="`echo ${FUNCSNAME[$IDX]} | sed -e 's/</$<$/g' -e 's/>/$>$/g'`"
    echo "\subsubsection{$ESCAPEDFUNCNAME}"
    echo "\addefinename{${FUNCSNAME[$IDX]}: ${FUNCSPARAMETER[$IDX]}}"
    echo "$ESCAPEDFUNCNAME is part of \adthistype."
    echo "\begin{adsignature}"
    echo " \item"
    echo ""
    echo " \begin{adsnippet}"
#    if [ -z "${FUNCSCONDITION[$IDX]}" ]
#    then
#      SNIPPETLINE="${FUNCSNAME[$IDX]}: ${FUNCSPARAMETER[$IDX]}"
#      SNIPPETTABBING=""
#      print_snippet_line
#    else
#      echo "${FUNCSCONDITION[$IDX]} {"
#      SNIPPETLINE="${FUNCSNAME[$IDX]}: ${FUNCSPARAMETER[$IDX]}"
#      SNIPPETTABBING="  "
#      print_snippet_line
#      echo "}"
#    fi
    if [ -z "${FUNCSCONDITION[$IDX]}" ]
    then
      LINE="${FUNCSNAME[$IDX]}: ${FUNCSPARAMETER[$IDX]}"
    else
      LINE="${FUNCSCONDITION[$IDX]} { ${FUNCSNAME[$IDX]}: ${FUNCSPARAMETER[$IDX]} }"
    fi
    CURCOL=0
    IDENTATION=""
    IDENTATIONCOLS=0
    print_aldor_expression
    echo " \end{adsnippet}"
    echo "\end{adsignature}"
    echo -e ${FUNCSDOCUMENTATION[$IDX]}
  done
fi

function check_expfile()
{
  cat "$EXPFILE" | grep -v -e "\adtype{$EXPORTEDTYPE}" >"$EXPFILE"
  if [ "`grep -c -e '\adtype{' $EXPFILE`" = "0" ]
  then
    rm "$EXPFILE"
  fi
}

#removing this type from all exported statements
EXPFILES=`find $DOCLIBRARYDOCDIR -name '*.exp' -printf '%h/%f ' | sed -e 's/ $//'`
if [ ! -z "$EXPFILES" ]
then
  if [ "`echo $EXPFILES | grep -c -e ' '`" = "0" ]
  then
    #just one file
    EXPFILE=$EXPFILES
    check_expfile
  else
    for EXPFILE in `grep -c -e "\adtype{$EXPORTEDTYPE}" $EXPFILES | grep -v -e ":0" | sed -e 's/:[0-9]*$//' `
    do
      check_expfile
    done
  fi
fi

#adding this type to the categories it is exporting
if [ $TYPES -gt 0 ] 
then
  FULLLIBSRCDIR="`echo $(pwd) | sed -e \"s|$DOCDIR/$SRCDIR|$LIBDIR/$SRCDIR|\" ` "
  for IDX in `seq 0 $(($TYPES-1))`
  do
    if [ "${TYPESNAME[$IDX]}" != "export" ]
    then
      EXPFILE="`grep \"PROVIDEDTYPE: ${TYPESNAME[$IDX]} -\" ../../$LIBDIR/$SRCDIR/$DEPENDENCYFILE | cut -d '-' -f 2 | sed -e \"s| $PROJECTROOT/$LIBDIR/$SRCDIR/|$DOCLIBRARYDOCDIR/|\" -e 's/.as$/.exp/'`"
      if [ ! -z "$EXPFILE" ]
      then
	# ${TYPESNAME[$IDX]} is exported by \projectname and can therefore be found.
	if [ ! -e "$EXPFILE" ]
	then
	  if [ ! -e "`dirname $EXPFILE`" ]
	  then
	    mkdir -p `dirname $EXPFILE`
	  fi
	  echo "\begin{adexportedby}"    >"$EXPFILE" 
	  echo "\item \adtype{$EXPORTEDTYPE}" >>"$EXPFILE" 
	  echo "\end{adexportedby}"     >>"$EXPFILE" 
	else
	  echo "\item \adtype{$EXPORTEDTYPE}" | cat "$EXPFILE" - | grep -e "\adtype{" | sort >"$EXPFILE"
	  echo "\end{adexportedby}" >>"$EXPFILE" 
	  echo "\begin{adexportedby}" | cat - "$EXPFILE" | cat >"$EXPFILE" 
	fi
      fi
    fi
  done
fi
exit 0
