#!/bin/sh

##########################################################################
#                                                                        #
#                        generate_testsuite.sh                           #
#                                                                        #
##########################################################################
#                                                                        #
# A script for generating a source file for an AldorUnit testsuite       #
#                                                                        #
##########################################################################
#                                                                        #
# Copyright (C) 2005  Research Institute for Symbolic Computation, J.    #
# Kepler University, Linz, Austria                                       #
#                                                                        #
# Written by Christian Aistleitner                                       #
#                                                                        #
# This program is free software; you can redistribute it and/or          #
# modify it under the terms of the GNU General Public License version 2, #
# as published by the Free Software Foundation.                          #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program; if not, write to the Free Software            #
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,             #
# MA  02110-1301, USA.                                                   #
#                                                                        #
##########################################################################

ERROROK=0
ERRORUNDEFINEDERROR=100
ERRORARGUMENTS=$(($ERRORUNDEFINEDERROR+1))
ERRORTYPENOTFOUND=$(($ERRORUNDEFINEDERROR+2))
ERRORAMBIGUOUSTYPE=$(($ERRORUNDEFINEDERROR+3))


function print_usage() {
cat <<HEREDOC

generate_testsuite.sh
----------------------

Usage:

  generate_testsuite.sh PROJECTNAME


Parameters:

  PROJECTNAME -  the name of the current project
                     (e.g.: CharSet)

Description:

This script generates an Aldor source code for an AldorUnit testsuite
HEREDOC
}

if [ $# -ne 1 ]
then
  print_usage
  exit $ERRORARGUMENTS
fi
PROJECTNAME=$1

## arguments are ok #################################################

CWD=`pwd`

SCRIPTPATH=`dirname "$0"`
SRCFILES=`find -follow -type f | grep -e \.as\$ | grep -v \.signatures\.as\$ | grep -v ^./tst_suit\.as | grep -v \./testcases.as`
for FILE in $SRCFILES
do
  TYPE=`${SCRIPTPATH}/get_type_for_file.sh "$FILE"`
  if [ $? -eq 0 ]
  then
    if [ `grep -c ^${TYPE}[\ \t]*: $FILE` -ge 1 ]
    then
      NEXTIDX=${#FILES[@]}
      FILES[$NEXTIDX]=$FILE
      TYPES[$NEXTIDX]=$TYPE
    fi
  else
    echo "cannot resolve type for $FILE" >&2
    exit 1
  fi
done
if [ $((${#FILES[@]}-1)) -lt 0 ]
then
  echo "No test cases found!" >&2
  exit 1
fi
INDICES=`seq 0 $((${#FILES[@]}-1))`

#--------------------------------------
cat <<ALDORCODE
-------------------------------------------------------------------------
--
-- automatically generated testsuite for $PROJECTNAME
--
-------------------------------------------------------------------------

#include "aldorunit"
#include "testcases"

-- Using DoubleEndedList if Algebra is available.
-- Defaulting to List otherwise. In order to keep to order of tests
-- for Algebra and non-Algebra setups consistent, List's append! is
-- used instead of insert!
#if LibraryAldorUnitRequiresLibraryAlgebra
  macro DELIST==DoubleEndedList;
  macro DELISTEMPTY==empty();
  macro DELISTAPPEND==concat!;
  macro DELISTFIRSTCELL( list )==firstCell list;
#else
  macro DELIST==List;
  macro DELISTEMPTY==empty;
  macro DELISTAPPEND==append!;
  macro DELISTFIRSTCELL( list )==list;
#endif

TestSuite : with {

ALDORCODE
#--------------------------------------


for IDX in $INDICES
do
  TYPE=${TYPES[$IDX]}
  echo "  runTestCase$TYPE: () -> Integer;"
done


#--------------------------------------
cat <<ALDORCODE

} == add {


ALDORCODE
#--------------------------------------


for IDX in $INDICES
do
  TYPE=${TYPES[$IDX]}
  FILE=${FILES[$IDX]}


#--------------------------------------
cat <<ALDORCODE
    -- tests for $TYPE (file: $FILE) ---------------------------------------------
 
    runTestCase$TYPE():Integer == {
        import from $TYPE;

        import from TestCaseTools;

        local testsList : DELIST NamedTest;

        testsList := DELISTEMPTY;

ALDORCODE
#--------------------------------------


IFPATTERN="^#if"
ELSEPATTERN="^#else"
ENDIFPATTERN="^#endif"
SIGNATURETESTPATTERN="[\ \t]*test\([a-zA-Z0-9\?\!]*\)[\ \t]*([\ \t]*)[\ \t]*:[\ \t]*([\ \t]*)[\ \t]*->[\ \t]*([\ \t]*)[\ \t]*;"
SIGNATURETESTPATTERN="[\ \t]*test\([a-zA-Z0-9\?\!]*\)[\ \t]*:[\ \t]*([\ \t]*)[\ \t]*->[\ \t]*([\ \t]*)[\ \t]*;"
SRCTESTPATTERN="test\([a-zA-Z0-9\?\!]*\)[\ \t]*([\ \t]*)[\ \t]*:[\ \t]*([\ \t]*)[\ \t]*==.*"
SEDREPLACEPATTERN="        testsList := DELISTAPPEND( testsList, generateNamedTest( \"\1\", test\1 ) );"

    
    #the grepping of the source file is not necessary, since a
    #signature file is assumed
    #grep -e "$IFPATTERN\|$ELSEPATTERN\|$ENDIFPATTERN\|$SRCTESTPATTERN" "$FILE" | sed -e "s/$SRCTESTPATTERN/$SEDREPLACEPATTERN/"
    for SIGNATUREFILE in `grep ^\#include.*signatures\.as\" "$FILE" | sed -e "s/[^\"]*\"\([^\"]*\)\"/\1/"`
    do
      SIGNATUREFILE="`dirname \"$FILE\"`/$SIGNATUREFILE"
      grep -e "$IFPATTERN\|$ELSEPATTERN\|$ENDIFPATTERN\|$SIGNATURETESTPATTERN" $SIGNATUREFILE | sed -e "s/$SIGNATURETESTPATTERN/$SEDREPLACEPATTERN/"
      
    done
#--------------------------------------
cat <<ALDORCODE

        processTestCase("$TYPE", setUp, DELISTFIRSTCELL testsList, tearDown);

    }

ALDORCODE
#--------------------------------------


done

cat <<ALDORCODE
}


-------------------------------------


ALDORCODE


#--------------------------------------

 
cat <<ALDORCODE

import from TestSuite;
import from TestSuiteTools;
import from NamedTestCase;
import from List NamedTestCase;

runTestSuite( "TestSuite", "$PROJECTNAME", [
ALDORCODE
COMMA=" "
for IDX in $INDICES
do
  TYPE=${TYPES[$IDX]}
  echo "    $COMMA generateNamedTestCase( \"runTestCase$TYPE\" , runTestCase$TYPE ) "
  COMMA=","
done
cat <<ALDORCODE
  ]);
ALDORCODE
#--------------------------------------
